/*
 * Copyright 2007 UbiCollab.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ubicollab.core.sdm;

import java.io.IOException;
import java.util.Collection;

import org.osgi.framework.BundleException;
import org.ubicollab.core.sm.model.Space;
import org.ubicollab.service.ServiceProxy;
import org.ubicollab.service.ServiceType;

/**
 * An interface for UbiCollab Service Domain Manager. Provides management functionality 
 * for services/bundles through web service calls.
 * 
 * TODO: Rename to ServiceDomainManager, ref Ubicollab whitepaper. Rename when coupling
 * with other platform services has been clarified.   
 * 
 * @author Kim Steve Johansen
 * @version 0.9
 */
public interface DomainManager {

	/**
	 * Gets the state of the service proxy provided by the bundle identified by bundleID.
	 * 
	 * @param bundleID the bundle ID for the bundle providing the proxy service
	 * 
	 * @return the bundle state
	 */
	String getServiceStateAsText(long bundleId);
	
	/**
	 * Gets all registered information about the specified service.
	 * <p>
	 * The xml-structure used to represent the returned service is:<br/>
	 * <pre>
	 *   &lt;?xml version='1.0'?&gt;
	 *      &lt;Service&gt;
	 *          &lt;Service ID=&quot;int&quot; 
	 *            ServiceUri=&quot;String&quot; 
	 *            Name=&quot;String&quot; 
	 *            Space=&quot;String&quot;
	 *            State=&quot;String&quot; 
	 *            DescriptionUri=&quot;String&quot; 
	 *            Type=&quot;String&quot; 
	 *            Description=&quot;String&quot;/&gt;<br/> 
	 *      &lt;/Service&gt;
	 * </pre>
	 * @param bundleID the bundle ID the bundle is registered with.
	 * 
	 * @return An xml-encoded <code>String</code> containg information about the service,
	 *         or "INVALID BUNDLE ID" if invalid bundle id is provided.
	 */
	String getInfoAboutService(long bundleID);
	
	/**
	 * Read from the log and return selected log items in a XML-formatted string.
	 * <p>
	 * The xml-structure used to represent the returned log is:<br/>
	 * <pre>
	 *   &lt;?xml version='1.0'?&gt;
	 *      &lt;Log&gt;
	 *          &lt;Logitem Date=&quot;dd.mm.yyyy&quot; 
	 *            Time=&quot;hh:mm:ss&quot; 
	 *            Level=&quot;String&quot; 
	 *            Message=&quot;String&quot;/&gt;<br/> 
	 *          ... more log items ...<br/>
	 *      &lt;/Log&gt;
	 * </pre>
	 * 
	 * @param logLevel The log level to include in the report. Valid arguments are:
	 *      'DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL' and 'ALL' for all types.
	 * @param numItems The number of log items to include in the report
	 * 
	 * @return the <code>String</code> containg the resulting XML
	 * 
	 * @deprecated The current implementation of DomainManager does not use logs.
	 */
	String getLogItems(int logLevel, int numItems);

	/**
	 * @return A Collection of ServiceProxy-s that are in the same Space as the current space
	 * of SpaceManager. 
	 */
	Collection<ServiceProxy> getNearbyServices();
	
	/**
	 * @param space The Space to retrieve ServiceProxy-s from.
	 * @return A Collection of ServiceProxy-s that are in the given Space.
	 */
	Collection<ServiceProxy> getServicesInSpace(Space space);
	
	/**
	 * @param type The ServiceType of the ServiceProxy-s to retrieve.
	 * @return A Collection of ServiceProxy-s with the given ServiceType.
	 */
	Collection<ServiceProxy> getServicesWithType(ServiceType type);

	/**
	 * @return The number of ServiceProxy-s registered.
	 */
	int getServiceCount();
	
	/**
	 * Return a list of the services which have been installed on this UbiNode through 
	 * the Service Domain Manager. 
	 * <p>
	 * The xml-structure used to represent the returned service list is:<br/>
	 * <pre>
	 *   &lt;?xml version='1.0'?&gt;
	 *      &lt;Servicelist&gt;
	 *          &lt;Service ID=&quot;int&quot; 
	 *            ServiceUri=&quot;String&quot; 
	 *            Name=&quot;String&quot; 
	 *            Space=&quot;Space&quot;
	 *            State=&quot;String&quot; 
	 *            DescriptionUri=&quot;String&quot; 
	 *            Type=&quot;String&quot; 
	 *            Description=&quot;String&quot;/&gt;<br/> 
	 *            ... more services ...<br/>
	 *      &lt;/Servicelist&gt;
	 * </pre>
	 * 
	 * @return The method returns a <code>String</code> with a XML-document 
	 *         which contains all installed services.    
	 */
	String getServiceListAsXml();
	
	/**
	 * @param bundleId The BundleId of the ServiceProxy to get.
	 * @return The ServiceProxy with the given BundleId.
	 */
	ServiceProxy getServices(long bundleId);
	
	/**
	 * @return All ServiceProxy-s registered with the DomainManager. 
	 */
	Collection<ServiceProxy> getServices();

	/**
	 * @param bundleId The bundleId of the ServiceProxy. 
	 * @return The state of the ServiceProxy. The state can be one of the states listed in
	 * org.osgi.framework.Bundle.
	 * 
	 * @see org.osgi.framework.Bundle
	 */
	int getServiceState(long bundleId);
	
	/**
	 * Returns a ServiceType with the given name, but only if a ServiceProxy registered with the
	 * DomainManager has a type with the given name.
	 * 
	 * @param nameOfTypeToView The name of the ServiceType.
	 * @return The ServiceType with the given name.
	 */
	ServiceType getServiceTypeByName(String nameOfTypeToView);

	/**
	 * Installs a service. Use this method if you have the URL to a service bundle.
	 * 
	 * @param friendlyName the friendly name of the service
	 * @param url The Url to the service bundle.
	 * 
	 * @return The ID of the installed bundle.
	 * 
	 * @throws BundleException	If there was an error installing the bundle. 
	 * @throws IOException		If there was an error installing the bundle.
	 */
	long installService(String url)
	throws BundleException, IOException;

	/**
	 * Removes the service provided by the bunde matching the provided bundleId.
	 * 
	 * @param bundleID the bundle ID of the service to remove
	 * 
	 * @return true, if service is removed, false othervise
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#removeService(java.lang.long)
	 */
	boolean removeService(long bundleID);
	
	/**
	 * Re-initializes the register of ServiceProxy-s.
	 */
	void rescanServices();

	/**
	 * Check if a given service exist.
	 * 
	 * @param bundleID the bundle ID
	 * 
	 * @return true, if service exists, false otherwise
	 */
	boolean serviceExist(long bundleId);

	/**
	 * Activates all services belonging to the space identified by the provided id. 
	 * Also provides the option to stop the other services.
	 * 
	 * @param space the Space to activate
	 * @param stopOtherServices true to stop the other service proxies, false othervise
	 * 
	 * @return <code>true<code> if all proxies adhering to the space is started, 
	 *         <code>false<code> othervise
	 */
	boolean setActiveSpace(Space space, boolean stopOtherServices);

	/**
	 * Method to start a service installed on the UbiNode.
	 * 
	 * @param bundleID the bundle ID for the bundle providing the service
	 * 
	 * @return <code>true<code> if the service is sucessfully started, 
	 *         <code>false<code> othervise
	 */
	boolean startService(long bundleID);

	/**
	 * Method to stop a service installed on the UbiNode
	 * 
	 * @param bundleID the bundle ID for the bundle providing the service
	 * @return <code>true<code> if the service is sucessfully stopped, 
	 *         <code>false<code> othervise
	 */
	boolean stopService(long bundleID);
	
	/**
	 * Method to update the properties of the service identified by serviceUri.
	 * 
	 * @param friendlyName The service friendly name
	 * @param description the textual description of the service
	 * @param space The space this service will be active in
	 * @param bundleId the bundle ID
	 * @throws IllegalArgumentException if Bundle with bundleId doesn't exist
	 * 
	 */
	void updateServiceProperties(long bundleId, String friendlyName, Space space, ServiceType type, String description);
	
}
