/*
 * Copyright 2007 UbiCollab.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ubicollab.core.sdm;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.BundleListener;
import org.ubicollab.core.sm.model.ISpaceManager;
import org.ubicollab.core.sm.model.Space;
import org.ubicollab.service.ServiceProxy;
import org.ubicollab.service.ServiceProxyFinder;
import org.ubicollab.service.ServiceType;


/**
 * The Class <code>DomainManagerImpl</code> implements the <code>DomainManager</code>
 * interface and provides the neccessary functionality for service management on
 * UbiNodes. The public members of this class will be accessible through a SOAP/Web-
 * Service interface.
 * 
 * @author Kim-Steve Johansen
 * @version 0.6
 * @since 0.1
 */
public class DomainManagerImpl implements DomainManager {
	/** Reference to the <code>BundleContext</code> this bundle runs in. */
	private BundleContext bundleContext;
	private ProxyList serviceList;
	private ISpaceManager spaceManager;
	private static DomainManager domainManagerInstance;
	
	/**
	 * This Constructor creates a new instance of <code>DomainManagerImpl</code>, and
	 * reads the service list from file.
	 * 
	 * @param bundleContext the <code>BundleContext</code> in which this bundle runs.
	 */
	public DomainManagerImpl(BundleContext bundleContext, ISpaceManager spaceManager) {
		domainManagerInstance = this;
		this.bundleContext = bundleContext;
		this.spaceManager = spaceManager;
		rescanServices();
	}

	public void rescanServices() {
		synchronized (this) {
			// Doing the initializations synchronized makes sure that no UbicollabService is
			// registered after serviceList has been created, but before BundleListener has started
			// listening for changes.
			initServiceList();
			initBundleListener(serviceList);
		}
	}

	private void initServiceList() {
		Debug.print("Scanning for Ubicollab services...");
		
		ArrayList<ServiceProxy> existingUbiServices =
			ServiceProxyFinder.searchForUbicollabServices(bundleContext);
		serviceList = new ProxyList(existingUbiServices, bundleContext);
		
		Debug.printServices(serviceList.getServices());
	}

	private void initBundleListener(ProxyList serviceList) {
		BundleListener ubiBundleListener = new UbicollabBundleListener(this, serviceList);
		bundleContext.addBundleListener(ubiBundleListener);
	}
	
	public String getServiceStateAsText(long bundleId) {
		return serviceList.getServiceStateAsText(bundleId);
	}
	
	public int getServiceState(long bundleId) {
		return bundleContext.getBundle(bundleId).getState();
	}
	
	public static DomainManager getInstance() {
		return domainManagerInstance;
	}

	/**
	 * Gets information about a service object, and returns it as serialized XML. 
	 * <p>
	 * The xml-structure used to represent the returned service is described in
	 * <code>DomainManager#getInfoAboutService</code>.
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#getInfoAboutService
	 * 
	 * @return A <code>String</code> containing a representation of the service list
	 *         or "INVALID BUNDLE ID" if bundle id does not exist.
	 */
	public String getInfoAboutService(long bundleID) {
		ServiceProxy serviceToGetInfoAbout = serviceList.getService(bundleID);
		
		String xml = GlobalSettings.XML_HEADER;
		xml += "<Service>";
		xml += serviceToGetInfoAbout.toXML( getServiceStateAsText( serviceToGetInfoAbout.getBundleId() ) );
		xml += "</Service>";
		
		return xml;
	}
	
	public Collection<ServiceProxy> getNearbyServices() {
		Space currentSpace = spaceManager.getCurrentSpace();
		return getServicesInSpace(currentSpace);
	}
	
	public Collection<ServiceProxy> getServicesInSpace(Space space) {
		Collection<ServiceProxy> servicesInTheSpace = new ArrayList<ServiceProxy>();
		
		Iterator<ServiceProxy> ubiServiceIterator = serviceList.getServices().iterator();
		
		while (ubiServiceIterator.hasNext()) {
			ServiceProxy ubiService = ubiServiceIterator.next();
			
			if (ubiService.getSpace().equals(space)) {
				servicesInTheSpace.add(ubiService);
			}
		}
		
		return servicesInTheSpace;
	}
	
	public Collection<ServiceProxy> getServicesWithType(ServiceType type) {
		Collection<ServiceProxy> servicesWithTheType = new ArrayList<ServiceProxy>();
		
		Iterator<ServiceProxy> ubiServiceIterator = serviceList.getServices().iterator();
		
		while (ubiServiceIterator.hasNext()) {
			ServiceProxy ubiService = ubiServiceIterator.next();
			
			if (ubiService.getType().equals(type)) {
				servicesWithTheType.add(ubiService);
			}
		}
		
		return servicesWithTheType;
	}
	

	/**
	 * Read from the log and return selected log items in a XML-formatted string.
	 * <p>
	 * The xml-structure used to represent the returned service is described in
	 * <code>DomainManager#getLogItems</code>.
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#getLogItems
	 * 
	 * @param logLevel The log level to include in the report.
	 * @param numItems The number of log items to include in the report
	 * 
	 * @return the string containg the resulting XML
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#getLogItems(int, int)
	 */
	public String getLogItems(int logLevel, int numItems) {
		Log log = Log.getInstance();
		return log.readFromLog(logLevel, numItems);
	}

	public int getServiceCount() {
		return serviceList.getServiceCount();
	}

	/**
	 * Searches for services installed on this UbiNode and returns handles to theese.
	 * <p>
	 * The xml-structure used to represent the returned service is described in
	 * <code>DomainManager#getServiceList</code>.
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#getServiceList
	 * @return The method returns a list of installed services
	 */
	public String getServiceListAsXml() {
		String sList = null;
		Log log = Log.getInstance();
		Debug.print("getServiceListAsXml() - serviceList = " + serviceList);
		try {
			sList = serviceList.toXML();
			log.writeToLog("Servicelist fetched by client.", Log.INFO);
		} catch (Exception e) {
			log.writeToLog("Failed to fetch service list from database", Log.WARNING);
			e.printStackTrace();
		}
		Debug.print("getServiceListAsXml() - returning " + sList);
		return sList;
	}
	
	public ServiceProxy getServices(long bundleId) {
		return serviceList.getService(bundleId);
	}
	
	public Collection<ServiceProxy> getServices() {
		return serviceList.getServices();
	}

	public ServiceType getServiceTypeByName(String nameOfTypeToView) {
		Iterator<ServiceProxy> ubiServiceIterator = serviceList.getServices().iterator();
		
		while (ubiServiceIterator.hasNext()) {
			ServiceProxy ubiService = ubiServiceIterator.next();
			ServiceType currentType = ubiService.getType();
			
			if (currentType.getName().equals(nameOfTypeToView))
				return currentType;
		}
		
		throw new RuntimeException("Type not found");
	}

	public long installService(String bundleSourceUrl)
	throws BundleException, IOException
	{
		Log log = Log.getInstance();
		log.writeToLog("Starting installation - 1: Installing bundle", Log.INFO);
		Bundle installedBundle = installBundle(bundleSourceUrl);
		// A BundleListener should listen for BundleEvent.INSTALLED, and continue with appropriate
		// actions. Currently, UbicollabBundleListener does this.
		
		return installedBundle.getBundleId();
	}

	/**
	 * Installs a bundle from the specified URL.
	 * 
	 * @param url						The URL of the bundle.
	 * @return							The bundle installed.
	 * @throws BundleException			If the provided stream cannot be read or the
	 * 									installation failed.
	 * @throws MalformedURLException	If the string specifies an unknown protocol.
	 * @throws IOException				If an I/O exception occurs.
	 */
	private Bundle installBundle(String url)
	throws BundleException, MalformedURLException, IOException {
		Debug.print("Installing bundle from: " + url);
		// The reason for separating local files from files from HTTP, is that
		// BundleContext.installBundle requires a InputStream for HTTP files while local
		// files require a String URL. The opposite creates a BundleException.
		// Environment: Android 2.1 emulator.
		if (urlIsHTTP(url)) {
			return installBundleFromHttp(url);
		} else {
			return installBundleFromLocalFile(url);
		}
	}

	/**
	 * Checks whether the URL starts with "http" or not.
	 * 
	 * @param url	The URL to check.
	 * @return		TRUE if the URL starts with "http".
	 */
	private boolean urlIsHTTP(String url) {
		return url.substring(0, 4).equals("http");
	}	
	
	/**
	 * Installs a Bundle from a file via HTTP. 
	 * 
	 * @param url						The URL to the file
	 * @return							The bundle installed.
	 * @throws BundleException			If the provided stream cannot be read or the
	 * 									installation failed.
	 * @throws MalformedURLException	If the string specifies an unknown protocol.
	 * @throws IOException				If an I/O exception occurs.
	 */
	private Bundle installBundleFromHttp(String url)
	throws MalformedURLException, IOException, BundleException {
		InputStream in = createInputStreamFromHttpUrl(url);
		Bundle installedBundle = bundleContext.installBundle(url, in);
		in.close();
		
		return installedBundle;
	}

	/**
	 * Opens and returns an <code>InputStream</code> for the contents of a given URL. 
	 * 
	 * @param url						The URL to read from.
	 * @return
	 * @throws MalformedURLException	If the string specifies an unknown protocol.
	 * @throws IOException				If an I/O exception occurs.
	 */
	private InputStream createInputStreamFromHttpUrl(String url)
			throws MalformedURLException, IOException {
		URL httpUrl = new URL(url);
		InputStream in = httpUrl.openStream();
		return in;
	}

	/**
	 * Installs a Bundle from a file on the local file system. 
	 * 
	 * @param url				The URL to the file
	 * @return					The bundle installed.
	 * @throws BundleException	If the provided stream cannot be read or the installation
	 * 							failed.
	 */
	private Bundle installBundleFromLocalFile(String url) throws BundleException {
		Bundle installedBundle = bundleContext.installBundle(url);
		return installedBundle;
	}

	/**
	 * Removes the service provided by the bunde matching the provided bundleId.
	 * 
	 * @param bundleID the bundle ID of the service to remove
	 * 
	 * @return true, if service is removed, false othervise
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#removeService(java.lang.long)
	 */
	public boolean removeService(long bundleID) {
		Log log = Log.getInstance();
		boolean wasAbleToRemoveService;
		
		try {
			Bundle b = bundleContext.getBundle(bundleID);
			b.uninstall();
			
			wasAbleToRemoveService = true;
		} catch (Exception e) {
			log.writeToLog("Failed to delete service identified by bundleId"
					+ bundleID, Log.ERROR);
			e.printStackTrace();
			
			wasAbleToRemoveService = false;
		}
		
		return wasAbleToRemoveService;
		
	}

	/**
	 * Method to update the properties of the service identified by serviceUri.
	 * 
	 * @param space the space the service is placed in
	 * @param friendlyName The service friendly name
	 * @param type the type of service
	 * @param description the service description
	 * @param space The space this service will be active in
	 * @param bundleID The ID for the bundle which hosts the requested service
	 * 
	 * @return True if URI is found and update is OK, false otherwise.

	 * @throws IllegalArgumentException if bundleId is invalid
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#updateServiceProperties(long, String, String, int, String, String)
	 */
	public void updateServiceProperties(
			long bundleID, String friendlyName, Space space, ServiceType type, String description)
	throws IllegalArgumentException
	{
		serviceList.updateService(bundleID, friendlyName, space, type, description);
	}
	
	/**
	 * Check if a given service exist.
	 * 
	 * @param uri the service uri
	 * 
	 * @return true, if service exists, false otherwise
	 */
	public boolean serviceExist(long bundleId) {
		return serviceList.serviceExists(bundleId);
	}
	
	
	
	/**
	 * Activates all services belonging to the space identified by the provided id.
	 * 
	 * @param Space the Space to activate
	 * @param stopOtherServices true to stop the other service proxies, false othervise
	 * 
	 * @return true, if all proxies adhering to the space ID is started, false othervise.
	 */
	/*
	public boolean setActiveSpace(Space space, boolean stopOtherServices) {
		return serviceList.setActiveSpace(space, stopOtherServices, this);
	}
	*/
	
	/**
	 * Activates all services belonging to the space identified by the provided space.
	 * 
	 * @param Space the Space to activate
	 * @param stopOtherServices true to stop the other service proxies, false othervise
	 * 
	 * @return true, if all proxies adhering to the space is started, false othervise.
	 */
	// TODO: This doesn't work if service is turned off, because serviceList.getStartedServces()
	// doesn't reveal services that are turned off (I renamed the method to this when I found out).  
	public boolean setActiveSpace(Space space, boolean stopOtherServices)
	{
		Debug.print("setActiveSpace(" + space + ")");
		// Check the Space of all registered services and search for a match with the given Space.
		// If a service is found with the same space, then start it.
		boolean allServicesWereStarted = true;
		
		Collection<ServiceProxy> ubiServices = serviceList.getServices();
		
		Iterator<ServiceProxy> ubiServiceIterator = ubiServices.iterator();
		while (ubiServiceIterator.hasNext()) {
			ServiceProxy ubiService = ubiServiceIterator.next();
			
			if (ubiService.getSpace().equals(space)) {
				boolean serviceWasStarted = startService(ubiService.getBundleId()); 
				allServicesWereStarted &= serviceWasStarted;
			} else {
				if (stopOtherServices) {
					stopService(ubiService.getBundleId());
				}
			}
		}
		
		return allServicesWereStarted;
	}
	
	/**
	 * Method to start a service installed on the UbiNode.
	 * 
	 * @param bundleId the bundle ID
	 * 
	 * @return True if the service is sucessfully started
	 */
	public boolean startService(long bundleId) {
		boolean serviceWasStarted = false;
		
		Bundle bundleToStart = bundleContext.getBundle(bundleId);
		try {
			bundleToStart.start();
			serviceWasStarted = true;
		} catch (BundleException e) {
			Log log = Log.getInstance();
			log.writeToLog("Failed to start service with ID=" + bundleId, Log.WARNING);
			e.printStackTrace();
		}

		return serviceWasStarted;
	}

	/**
	 * Method to stop a service installed on the UbiNode
	 * 
	 * @param bundleID the bundle ID
	 * @return True if the service is sucessfully started
	 */
	public boolean stopService(long bundleId) {
		boolean serviceWasStopped = false;
		
		Bundle bundleToStart = bundleContext.getBundle(bundleId);
		try {
			bundleToStart.stop();
			serviceWasStopped = true;
		} catch (BundleException e) {
			Log log = Log.getInstance();
			log.writeToLog("Failed to stop service with ID=" + bundleId, Log.WARNING);
			e.printStackTrace();
		}

		return serviceWasStopped;
	}
	
}