/*
 * Copyright 2007 UbiCollab.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ubicollab.core.sdm;

// TODO legg til add og remove p servicelist, makes sense at det er der og dm kun har install og remove

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.ubicollab.core.sm.model.Space;
import org.ubicollab.service.ServiceProxy;
import org.ubicollab.service.ServiceType;


/**
 * The Class <code>ServiceList</code> is used to represent all services which
 * have been registered with the Service Domain Manager. 
 * <p>
 * The class functionality neccessary to handle <code>Service</code> management tasks.
 * 
 * The list of services created by using factory method createServiceList, and by calling
 * rescanServices, for instance from a org.osgi.framework.ServiceListener (when services are
 * registered, unregistered or modified.) In that way, the list will always be consistent with
 * the actual existent ServiceProxy-s.
 * 
 * @author Kim-Steve Johansen
 * @version 0.5
 * @since 0.1
 */
public class ProxyList {
	private ProxyListContainer listContainer;
	private BundleContext bundleContext;

	ProxyList(ArrayList<ServiceProxy> existingServicesArrList, BundleContext bundleContext) {
		listContainer = createServiceListContainerFrom(existingServicesArrList);
		this.bundleContext = bundleContext;
	}

	private ProxyListContainer createServiceListContainerFrom(
			ArrayList<ServiceProxy> existingServices)
	{
		ProxyListContainer newListContainer = new ProxyListContainer(); 
		
		Iterator<ServiceProxy> iter = existingServices.iterator();
		while (iter.hasNext()) {
			ServiceProxy ubiService = iter.next();
			newListContainer.put(ubiService);
		}
		
		return newListContainer;
	}

	/**
	 * Check if a given service exist (is member of the service list).
	 * 
	 * @param uri the service uri
	 * 
	 * @return true, if service exists, false otherwise
	 */
	boolean serviceExists(long bundleId) {
		return listContainer.serviceExists(bundleId);
	}

	/**
	 * Returns the UbicollabService with the given bundleId.
	 * 
	 * @param bundleID The bundle ID of the UbicollabService.
	 * 
	 * @return The UbicollabService with the given bundleId, or null if none found.
	 * 
	 * @throws IllegalArgumentException if bundleId wasn't found
	 */
	ServiceProxy getService(long bundleId)
	throws IllegalArgumentException
	{
		return listContainer.get(bundleId);
	}
	
	Collection<ServiceProxy> getServices() {
		return listContainer.getServices();
	}

	/**
	 * Method to serialize the service list to XML. 
	 * <p>
	 * The xml-structure used to represent the returned service is described in
	 * <code>DomainManager#getServiceList</code>.
	 * 
	 * @see no.ubicollab.osgi.service.domainmanager.DomainManager#getServiceList
	 * 
	 * @return A <code>String</code> containing a representation of the service list
	 */
	String toXML() {
		Debug.print("ServiceList.toXML");
		String xml = GlobalSettings.XML_HEADER;
		xml += "<Servicelist>";
		
		Collection<ServiceProxy> ubiServices = getServices();
		Iterator<ServiceProxy> iter = ubiServices.iterator();
		while (iter.hasNext()) {
			ServiceProxy ubiService = iter.next();
			xml += ubiService.toXML( getServiceStateAsText( ubiService.getBundleId() ) );
		}
		/*
		for (int i=0;i<serviceList.size();i++) {
			ubiService = serviceList.get(i);
			xml += ubiService.toXML( getBundleStateAsText( ubiService.getBundleID() ) ); 
		}
		*/
		xml += "</Servicelist>";
		return xml;
	}
	

	/**
	 * Gets the state of the service proxy provided by the bundle identified by bundleID.
	 * 
	 * @param bundleID the bundle ID for the bundle providing the proxy service
	 * 
	 * @return the bundle state
	 */
	String getServiceStateAsText(long bundleID) {
		Bundle bundle = bundleContext.getBundle( bundleID );
		String bundleState = "unknown";
		switch (bundle.getState()) {
			case Bundle.ACTIVE: bundleState = "Started"; break;
			case Bundle.INSTALLED: bundleState = "Installed"; break;
			case Bundle.RESOLVED: bundleState = "Stopped"; break;
			case Bundle.STARTING: bundleState = "Starting"; break;
			case Bundle.STOPPING: bundleState = "Stopping"; break;
			case Bundle.UNINSTALLED: bundleState = "Uninstalled"; break;
		}
		return bundleState;
	}

	/**
	 * Method to update the properties of the service identified by serviceUri
	 * 
	 * @param bundleId The id for the bundle which hosts the service
	 * @param friendlyName The service friendly name
	 * @param space The space this service will be active in
	 * 
	 * @throws IllegalArgumentException if bundleId is invalid
	 */
	void updateService(long bundleId, String friendlyName, Space space, ServiceType type,
			String description)
	throws IllegalArgumentException
	{
		ServiceProxy ubiService = getService(bundleId);
		ubiService.setName(friendlyName);
		ubiService.setSpace(space);
		ubiService.setType(type);
		ubiService.setDescription(description);
	}

	int getServiceCount() {
		return getServices().size();
	}
	
	void put(ServiceProxy ubiService) {
		listContainer.put(ubiService);
	}

	ServiceProxy get(long bundleId) {
		return listContainer.get(bundleId);
	}
	
	ServiceProxy remove(long bundleId) throws IllegalArgumentException {
		return listContainer.remove(bundleId);
	}

	
	
}
