package org.ubicollab.service;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.util.tracker.ServiceTracker;
import org.osgi.util.tracker.ServiceTrackerCustomizer;
import org.ubicollab.service.exception.InvalidSearchException;
import org.ubicollab.service.exception.InvalidServiceLinkClassException;
import org.ubicollab.service.exception.InvalidServiceLinkException;
import org.ubicollab.service.exception.NoServiceLinkFoundException;
import org.ubicollab.service.exception.TooManyServiceLinksFoundException;

public class ServiceLinkGetter {

	private static Bundle bundle;

	public static ServiceLink getOsgiService(BundleContext bundleContext, String className)
	throws InvalidServiceLinkException
	{
		ServiceTracker serviceListTracker = getServiceTracker(bundleContext, className);
		Object serviceInstance = serviceListTracker.getService();
		Debug.print("getOsgiService result: " + serviceInstance);
		
		if (serviceInstance instanceof ServiceLink) {
			return (ServiceLink) serviceInstance; 
		} else {
			throw new InvalidServiceLinkException("The OSGi service received using class name " +
					className + " was not a ServiceLink.");
		}
	}

	private static ServiceTracker getServiceTracker(BundleContext bundleContext, String className) {
		ServiceTracker serviceListTracker = new ServiceTracker(bundleContext, className, null);
		serviceListTracker.open();
		
		return serviceListTracker;
	}
	

	/**
	 * Retrieves the ServiceLink of a Bbundle.
	 * 
	 * @param bundle The Bundle in which to search for a ServiceLink.
	 * @return The ServiceLink found.
	 * @throws NoServiceLinkFoundException if no ServiceLink was found. 
	 */
	public static ServiceLink searchForServiceLink(Bundle bundle)
	throws NoServiceLinkFoundException
	{
		ServiceLinkGetter.bundle = bundle;
		
		try {
			return tryToSearchForServiceLink();
		} catch (InvalidSyntaxException e) {
			e.printStackTrace();
			throw new NoServiceLinkFoundException(e.getMessage());
		} catch (InvalidSearchException e) {
			e.printStackTrace();
			throw new NoServiceLinkFoundException(e.getMessage());
		} 
	}

	private static ServiceLink tryToSearchForServiceLink()
	throws InvalidSyntaxException, TooManyServiceLinksFoundException,
	InvalidServiceLinkClassException
	{
		ServiceTracker serviceLinkTracker = createServiceLinkTracker();
		serviceLinkTracker.open(); // only search services within bundle
		
		checkSizeOfServiceLinkTracker(serviceLinkTracker);
		
		Object serviceLinkObj = serviceLinkTracker.getService();
		if (! (serviceLinkObj instanceof ServiceLink))
			throw new InvalidServiceLinkClassException(
					"The ServiceLink found was not a ServiceLink instance.");
		
		ServiceLink serviceLinkFound = (ServiceLink) serviceLinkObj;
		return serviceLinkFound;
	}

	private static ServiceTracker createServiceLinkTracker()
			throws InvalidSyntaxException {
		BundleContext bundleContext = bundle.getBundleContext();
		Filter serviceLinkSearchFilter = createServiceLinkSearchFilter();
		ServiceTrackerCustomizer customizer = null;
		
		ServiceTracker serviceLinkTracker =
			new ServiceTracker(bundleContext, serviceLinkSearchFilter, customizer);
		return serviceLinkTracker;
	}

	private static void checkSizeOfServiceLinkTracker(ServiceTracker serviceLinkTracker) {
		if (serviceLinkTracker.size() > 1)
			throw new TooManyServiceLinksFoundException(
					"Only one ServiceLink is supposed to be offered by a UbicollabService." +
					"Services found: " + serviceLinkTracker.size());
	}
	
	private static Filter createServiceLinkSearchFilter() throws InvalidSyntaxException {
		// (&(bundleId=num)(property=val))
		long bundleId = bundle.getBundleId();
		
		String serviceLinkSearchFilterStr =
			"(&" +
				"(" +
					ServiceLinkRegisterer.UBICOLLAB_BUNDLE_ID_PROPERTY_NAME +
					"=" +
					bundleId +
				")" +
				"(" +
					ServiceLinkRegisterer.OSGI_SERVICES_ID_PROPERTY_NAME +
					"=" +
					ServiceLinkRegisterer.OSGI_SERVICES_ID_VALUE +
				")" +
			")";
		Filter serviceLinkSearchFilter = FrameworkUtil.createFilter(serviceLinkSearchFilterStr);
		
		return serviceLinkSearchFilter;
	}


}
