package org.ubicollab.testing.sdm.junit;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collection;
import java.util.Iterator;

import junit.framework.TestCase;

import org.osgi.framework.BundleException;
import org.ubicollab.core.sdm.DomainManager;
import org.ubicollab.core.sdm.GlobalSettings;
import org.ubicollab.core.sm.model.Space;
import org.ubicollab.service.ServiceProxy;
import org.ubicollab.service.ServiceType;

/**
 * JUnit test for a ServiceListService.
 * 
 * - A note about Constructor parameter and independence from OSGi: 
 * Note that this test needs to be provided with an existing ServiceListService instance.
 * It would perhaps be optimal to create a new one in this test to ensure a clean state
 * of the instance. However, a ServiceListService needs a valid BundleContext from the
 * OSGi framework. To keep the test clean and independent from how a ServiceListService
 * gets this BundleContext (and thus independent from OSGi), this test must be provided
 * with an existing ServiceListService instance. Alternatively, this could create a new
 * ServiceListService, but this would require access to a OSGi bundle's BundleContext.
 * 
 * @author Yngvar Kristiansen
 */

// Tests ignored:
// getLogItems: The current implementation of DomainManager does not use logs.

public class DomainManagerTest extends TestCase {
	// Bundles required for testSetActiveSpace
	private static final String urlToHelloWorldbundle2 =
		"http://yngvark.com/files/HelloWorldUbicollab2.jar";
	private static final String urlToHelloWorldbundle3 =
		"http://yngvark.com/files/HelloWorldUbicollab3.jar";
	
	private DomainManager domainManager;
	private long helloWorldBundleId;

	public DomainManagerTest() {
		helloWorldBundleId = Activator.getInstance().getHelloWorldBundle(); 
		domainManager = Activator.domainManager;
	}
	
	public void testInstallAndRemoveService() throws BundleException {
		Debug.print("testInstallAndRemoveService");
		try {
			tryToInstallHelloWorldBundle();
		} catch (IOException e) {
			e.printStackTrace();
			assertTrue("Invalid URL: " + urlToHelloWorldbundle2, false);
		}
	}

	private void tryToInstallHelloWorldBundle() throws BundleException, IOException {
		long installedBundleId = installHelloWorldBundle(urlToHelloWorldbundle2);
		domainManager.startService(installedBundleId);
		
		boolean serviceExists = domainManager.serviceExist(installedBundleId);
		assertTrue("Service does not exist according to DomainManager.", serviceExists);

		boolean uninstallSucceeded = domainManager.removeService(installedBundleId);
		assertTrue("removeService did not succeed", uninstallSucceeded);
	}
	
	private long installHelloWorldBundle(String urlToHelloWorldbundle)
	throws BundleException, IOException
	{
		long installedBundleId = domainManager.installService(urlToHelloWorldbundle);
		domainManager.startService(installedBundleId);
		
		return installedBundleId;
	}
	
	public void testGetInfoAboutService()
	throws MalformedURLException, BundleException, InterruptedException, IOException
	{
		Debug.print("testGetInfoAboutService");
		
		String info = domainManager.getInfoAboutService(helloWorldBundleId);
		Debug.print("HelloWorldBundle info: " + info);
		
		stringShouldNotBeEmpty(info, "getInfoAboutService() result was empty");
		resultShouldBeServiceXML(info, "getInfoAboutService() did not return valid XML.");
	}

	private void stringShouldNotBeEmpty(String result, String errorMsg) {
		assertNotNull(errorMsg, result);
		boolean stringHasLengthOverZero = result.length() > 0;
		assertTrue(errorMsg, stringHasLengthOverZero);
	}
	
	private void resultShouldBeServiceXML(String xmlResult, String errorMsg) {
		checkXMLPrefix(xmlResult, "<Service>", errorMsg);
		checkXMLPostfix(xmlResult, "</Service>", errorMsg);
	}

	private void checkXMLPrefix(String xmlResult, String validXmlStart, String errorMsg) {
		String validPrefix = "";
		validPrefix += GlobalSettings.XML_HEADER;
		validPrefix += validXmlStart;
		
		String resultPrefix = xmlResult.substring(0, validPrefix.length());
		assertEquals(errorMsg, validPrefix, resultPrefix);
	}

	private void checkXMLPostfix(String xmlResult, String validPostfix, String errorMsg) {
		int beginIndex = xmlResult.lastIndexOf(validPostfix);
		String resultPostfix = xmlResult.substring(beginIndex);
		
		assertEquals(errorMsg, validPostfix, resultPostfix);		
	}

	public void testGetServices() {
		Debug.print("testGetServiceList");
		Collection<ServiceProxy> services = domainManager.getServices();
		
		Iterator<ServiceProxy> serviceIter = services.iterator();
		boolean helloWorldBundleFoundInServiceList = false;
		while (serviceIter.hasNext()) {
			if (serviceIter.next().getBundleId() == helloWorldBundleId)
				helloWorldBundleFoundInServiceList = true;
		}
			
		assertTrue(
				"HelloWorldBUndle wasn't found in getServices()",
				helloWorldBundleFoundInServiceList);
	}
	
	public void testGetServiceListAsXml() {
		Debug.print("testGetServiceListAsXml");
		
		String xmlServiceListResult = domainManager.getServiceListAsXml();
		
		assertNotNull("getServiceListAsXml() returned null", xmlServiceListResult);
		assertTrue("getServiceListAsXml().length() was <= 0", xmlServiceListResult.length() > 0);
		
		resultShouldBeServiceListXML(xmlServiceListResult, "getServiceListAsXml() did not return valid XML");
	}
	
	private void resultShouldBeServiceListXML(String xmlResult, String errorMsg) {
		checkXMLPrefix(xmlResult, "<Servicelist>", errorMsg);
		checkXMLPostfix(xmlResult, "</Servicelist>", errorMsg);
	}

	public void testUpdateServiceProperties() {
		Debug.print("testUpdateServiceProperties");
		
		String newFriendlyName = "HelloWorld Uno - changed by testUpdateServiceProperties.";
		Space newSpace = new Space();
		newSpace.setId(35);
		newSpace.setName("newSpaceName2323452");
		ServiceType newType = new ServiceType("newType239842983");
		
		String newDescription = "Another description - changed by testUpdateServiceProperties.";

		updateServicePropertiesOfHelloWorldBundle(
				newFriendlyName, newSpace, newType, newDescription
				);
	
		checkThatAllValuesAreUpdated(
				newFriendlyName, newSpace, newType, newDescription
				);
	}

	private void updateServicePropertiesOfHelloWorldBundle(String newFriendlyName, Space newSpace,
			ServiceType newType, String newDescription)
	{
		// long bundleID, String friendlyName, Space space, String type, String description)
		domainManager.updateServiceProperties(
				helloWorldBundleId,
				newFriendlyName,
				newSpace,
				newType,
				newDescription);
	}

	private void checkThatAllValuesAreUpdated(String newFriendlyName, Space newSpace,
			ServiceType newType, String newDescription)
	{
		ServiceProxy ubiService = domainManager.getServices(helloWorldBundleId);
		
		assertEquals("updateServiceProperties: newFriendlyName was not set propertly",
				newFriendlyName, ubiService.getName());
		assertEquals("updateServiceProperties: newSpace was not set propertly",
				newSpace, ubiService.getSpace());
		assertEquals("updateServiceProperties: newType was not set propertly",
				newType, ubiService.getType());
		assertEquals("updateServiceProperties: newDescription was not set propertly",
				newDescription, ubiService.getDescription());
	}

	public void testStartStopService() {
		Debug.print("testStartStopService");
		
		domainManager.stopService(helloWorldBundleId);
		bundleStateShouldBe("Stopped", helloWorldBundleId);
		
		domainManager.startService(helloWorldBundleId);
		bundleStateShouldBe("Started", helloWorldBundleId);
	}
	
	private void bundleStateShouldBe(String state, long bundleId) {
		String bundleState = domainManager.getServiceStateAsText(bundleId);
		assertEquals(state, bundleState);
	}
	
	public void testSetActiveSpace()
	throws MalformedURLException, BundleException, InterruptedException, IOException
	{
		Debug.print("testSetActiveSpace");
		
		// ------------------------ INITIALIZE TESTS ---------------------------------------
		
		// Install bundles
		long bundleIdHw2 = domainManager.installService(urlToHelloWorldbundle2);
		long bundleIdHw3 = domainManager.installService(urlToHelloWorldbundle3);

		// Start bundles
		domainManager.startService(bundleIdHw2);
		domainManager.startService(bundleIdHw3);

		// Get service intances
		ServiceProxy ubiService2 = domainManager.getServices(bundleIdHw2);
		ServiceProxy ubiService3 = domainManager.getServices(bundleIdHw3);

		// Create some test spaces
		Space joesFarm = new Space();
		joesFarm.setId(894196);
		joesFarm.setName("Joe's farm");
		
		Space evesMansion = new Space();
		evesMansion.setId(98437);
		evesMansion.setName("Eve's mansion");
		
		Space nowhere = new Space();
		nowhere.setId(4689);
		nowhere.setName("Nowhere");

		// Set service's space
		ubiService2.setSpace(joesFarm);
		ubiService3.setSpace(evesMansion);

		// ------------------------ TESTS ------------------------------------------------ 
		
		// 1st test: Stop the installed services, set space to nowhere, and check that the
		// services are still inactive
		Debug.print("testSetActiveSpace 1st test");
		domainManager.stopService(bundleIdHw2);
		domainManager.stopService(bundleIdHw3);
		
		domainManager.setActiveSpace(nowhere, false);
		
		// State of bundles should be stopped, because their space has not been set active. 
		bundleStateShouldBe("Stopped", bundleIdHw2);
		bundleStateShouldBe("Stopped", bundleIdHw3);
		
		// 2nd test: Set space of both services to Joe's farm, set active space to Joe's farm,
		// and then check that both services are active.
		Debug.print("testSetActiveSpace 2nd test");
		ubiService2.setSpace(joesFarm);
		ubiService3.setSpace(joesFarm);
		
		boolean allServicesWasStarted = domainManager.setActiveSpace(joesFarm, false);
		
		assertTrue("One or more services wasn't started when doing setActiveSpace",
				allServicesWasStarted);
		
		// State of bundle should be started.
		bundleStateShouldBe("Started", bundleIdHw2);
		bundleStateShouldBe("Started", bundleIdHw3);
		
		domainManager.removeService(bundleIdHw2);
		domainManager.removeService(bundleIdHw3);
	}
	

}
