package org.ubicollab.view.servlet;

import java.io.IOException;
import java.io.PrintWriter;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.osgi.framework.BundleContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.osgi.util.tracker.ServiceTracker;

@SuppressWarnings("serial")
public class GUIServlet extends HttpServlet { 
	private String servletAlias;
	private String servletBaseUrl;
	private URLHandler urlHandler;
	private LinkedHTMLPageFactory htmlPageFactory;
	
	public static final String PAGE_PARAM_NAME = "pageRequested10391683";

	// OSGi related
	private BundleContext bundleContext;
	private ServiceTracker httpTracker;
	private HttpService httpService = null;
	

	public GUIServlet(String servletAlias, BundleContext bundleContext) {
		this(servletAlias, bundleContext, null);
	}
	
	public GUIServlet(
			String servletAlias,
			BundleContext bundleContext,
			LinkedHTMLPageFactory htmlPageFactory)
	{
		this.servletAlias = servletAlias;
		this.bundleContext = bundleContext;
		initBaseUrl();
		initHtmlPageFactory(htmlPageFactory);
		this.urlHandler = new URLHandler(getBaseUrl());
	}
	

	protected void doGet(HttpServletRequest req, HttpServletResponse httpResponse) {
		doRequest(req, httpResponse);
	}
	
	private void doRequest(HttpServletRequest req, HttpServletResponse httpResponse) {
		try {
			String htmlResponse = getHtmlResponse(req);
			writeResponse(httpResponse, htmlResponse);
		} catch (Exception e) {
			// View all exceptions from the GUI so they don't go unnoticed (OSGi doesn't
			// show exceptions as standard).
			Debug.print("GUIServlet catched exception: " + e.getMessage());
			e.printStackTrace();
		}
	}

	protected void doPost(HttpServletRequest req, HttpServletResponse httpResponse) {
		doRequest(req, httpResponse);
	}
	

	private String getHtmlResponse(HttpServletRequest req) {
		HTMLPage responsePage = htmlPageFactory.getHtmlPage(req);
		String htmlResponse = responsePage.getHtml();
		return htmlResponse;
	}
	

	private void initBaseUrl() {
		String servletHttpPort = System.getProperty("org.osgi.service.http.port", "1149");
		servletBaseUrl = "http://localhost:" + servletHttpPort + "/" + servletAlias;
	}	

	private void initHtmlPageFactory(LinkedHTMLPageFactory htmlPageFactory) {
		if (htmlPageFactory == null) {
			this.htmlPageFactory = new LinkedHTMLPageFactory();
		} else {
			this.htmlPageFactory = htmlPageFactory;
		}
	}	

	public void addHtmlPage(LinkedHTMLTemplate htmlPage) {
		setHtmlPageRouting(htmlPage);
		setHtmlPageUrlHandler(htmlPage);
		setHtmlPageUrl(htmlPage);
	}

	private void setHtmlPageRouting(LinkedHTMLTemplate htmlPage) {
		htmlPageFactory.setRoutingFromPageIdToHtmlPage(htmlPage.getPageId(), htmlPage);
	}

	private void setHtmlPageUrlHandler(LinkedHTMLTemplate htmlPage) {
		htmlPage.setUrlHandler(urlHandler);
	}

	private void setHtmlPageUrl(LinkedHTMLTemplate htmlPage) {
		String pageUrl = "?" + PAGE_PARAM_NAME + "=" + htmlPage.getPageId();
		urlHandler.setPageUrl(htmlPage, pageUrl);
	}

	public String getServletAlias() {
		return servletAlias;
	}

	private String getServletAliasForHttpService() {
		return "/" + servletAlias;
	}

	public String getBaseUrl() {
		return servletBaseUrl;
	}

	private void writeResponse(HttpServletResponse httpResponse,
			String htmlResponse) {
		PrintWriter httpResponseWriter = getHttpResponseWriter(httpResponse);
		httpResponseWriter.print(htmlResponse);
		httpResponseWriter.close();
	}

	private PrintWriter getHttpResponseWriter(HttpServletResponse httpResponse) {
		httpResponse.setContentType("text/html");
		
		// Get the Writer AFTER setContentType - this ensures that
		// the obtained writer will property perform character set conversion.
		// Produce a HTML page according to the request.
		try {
			PrintWriter httpResponseWriter = httpResponse.getWriter();
			return httpResponseWriter;
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException("Could not get httpRseponseWriter.");
		}
		
	}

	public void registerHttpServlet() {
		// Create HttpService tracker
		httpTracker = new ServiceTracker(bundleContext, HttpService.class.getName(), null);
		httpTracker.open();
		httpService = (HttpService) httpTracker.getService();
		if (httpService != null) {
			// Register the servlet in the HTTP Service
			try {
				httpService.registerServlet(
						getServletAliasForHttpService(), this, null, null);
				Debug.print("Servlet registered with alias " + servletAlias);
			} catch (ServletException e) {
				Debug.print("ERROR: if the servlet's init method throws an exception," +
						"or the given servlet object has already been registered at a" +
						"different alias.");
				e.printStackTrace();
			} catch (NamespaceException e) {
				Debug.print("ERROR: Alias already in use!");
				e.printStackTrace();
			}
		} else {
			Debug.print("ERROR: httpService was null!!");
		}
	}	
	
	public void unregisterHttpServlet() {
		if (httpService != null) {
			try {
				httpService.unregister(getServletAliasForHttpService());
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			}
		}

		if (httpTracker != null) {
			httpTracker.close();
		}
	}
	
}
