<?php
/*
 * ------------------------------------------------------------------------------------------------
 * DESCRIPTION:
 * This class does two things (todo: clean up and refactor.):
 *
 * 1. IMAGE BROWSER FOR IMAGES IN A GIVEN DIRECTORY
 * Shows an HTML browser, that can browse through images in a given directory. The
 * requirement/limitation is that images has to be in sequence, for instance:
 * "img1.jpg", "img2.jpg", "img3.jpg", and so on.
 *
 * 2. REMOTE CONTROL OF IMAGE BROWSER
 * Gives the option to launch a browser on the server machine (the php/web server)
 * and show the same page as in point 1. Further, when the client hits the "next" or
 * previous button, the page is changed both on client and on the server. In this way,
 * it is possible for a client to remotely control the server, which can be useful for
 * showing and controlling a presentation from a client device.
 * ------------------------------------------------------------------------------------------------
 * DEPENDENCIES:
 * - An external browser in order to view the ImageBrowser output (html) on the server.
 * ------------------------------------------------------------------------------------------------
 * HOW TO RUN:
 * Given that a directory exists and contains images, using a numbered naming convention 
 * that can be configured), then you can this file with the following GET parameters
 * <url to this PHP>?imgDir=<name of image directory>&imgNum=<image number to show>&isClient=<true/false>
 *
 * Regarding parameter "isClient", there are two options:
 * 1. Set to "true" if you DO want to enable remote control (described above).
 * 2. Set to "false" if you DO NOT want this.
 *
 * For instance:
 * http://localhost/ImageBrowser.php?imgDir=out_k7bJicOd2F&imgNum=9&isClient=false
 *
 * ------------------------------------------------------------------------------------------------
 * TIPS:
 * - Open links, that would normally open in a new window, in the current tab/window:
 * The external browser is started using command <browser executable> <url>, for
 * instance firefox.exe http://localhost/ImageBrowser.php. You should set your browser
 * not to open a new tab when run this way. In Firefox, this is done by opening
 * config (in URL tab, write about:config), and changing setting
 * "browser.link.open_newwindow" to 1.
 *
 * See http://kb.mozillazine.org/Browser.link.open_newwindow for further details.
 * ------------------------------------------------------------------------------------------------
 *
 * Yngvar Kristiansen, 11.05.2010
*/

require_once("Config.php");
require_once("ErrorController.php");
require_once("HtmlPageBuilder.php");

class ImageBrowser {
	private $errorCtrl;
	private $htmlPageBuilder;

	private $imgPrefix = "img";
	private $imgSuffix = ".jpg";
	private $html;
	private $currentImageNum;
	private $imgDir;
	private $isClient;
	
	public function ImageBrowser($imgDir, $currentImageNum, $isClient) {
		$this->errorCtrl = ErrorController::getInstance();

		$this->setImgDir($imgDir);
		$this->setCurrentImageNum($currentImageNum);
		$this->setIsClient($isClient);
		
		$this->htmlPageBuilder = new HtmlPageBuilder();
		$this->checkThatExternalBrowserExists();
		$this->initImageBrowser();
	}
	
	private function setImgDir($imgDir) {
		if (! is_dir($imgDir)) {
			$this->errorCtrl->addError("Invalid image directory given");
		} else {
			$this->imgDir = $imgDir;
		}
	}
	
	private function setCurrentImageNum($currentImageNum) {
		$this->currentImageNum = $currentImageNum;
	}
		
	private function setIsClient($isClient) {
		$this->isClient = $isClient;
	}
	
	private function initImageBrowser() {
		$this->addCtrlButtons();
		$this->addImage();
	}
	
	private function addCtrlButtons() {
		if ($this->isClient()) {
			$this->addCtrlButtonsWithImageWidth(70);
		} else {
			$this->addCtrlButtonsWithImageWidth(40);
		}
	}
	
	private function addCtrlButtonsWithImageWidth($buttonWidth = 40) {
		$this->htmlPageBuilder->addHtml("\n<center>");
		if ($this->currentImageNum > 1) {
			$this->addPrevNextBtn($this->currentImageNum - 1, "<img src='img/Previous-icon.png' border='0' width='" . $buttonWidth . "' />");
		}
		if ($this->currentImageNum < $this->getImageCountOfImgDir()) {
			$this->addPrevNextBtn($this->currentImageNum + 1, "<img src='img/Next-icon.png' border='0' width='" . $buttonWidth . "' />");
		}
		$this->htmlPageBuilder->addHtml("</center>");
		$this->addBreak();
	}
	
	private function addPrevNextBtn($imgNum, $linkDescr) {
		$this->htmlPageBuilder->addHtml("<a href='ImageBrowser.php?imgDir=" . $this->imgDir . "&imgNum=" . $imgNum . "&isClient=" . $this->isClient . "'>$linkDescr</a> ");
	}
			
	private function getImageCountOfImgDir() {
		$currentImgNum = 1;
		while (file_exists($imgFile = $this->getImageUrl($currentImgNum))) {
			$currentImgNum++;
		}
		
		return $currentImgNum - 1;
	}
	
	private function getImageUrl($imgNum) {
		return $this->imgDir . "/" . $this->imgPrefix . $imgNum . $this->imgSuffix;
	}
	
	private function addImage() {
		$this->htmlPageBuilder->addHtml("<center><img src='" . $this->getImageUrl($this->currentImageNum) . "'></center>");
	}
	
	private function checkThatExternalBrowserExists() {
		$filepath = Config::FILEPATH_TO_BROWSER();
		if ($this->isStringSurroundedWithQuotes($filepath)) {
			$filepath = $this->removeStartAndEndCharFrom($filepath);
		}
		
		if (! file_exists($filepath)) {
			$this->errorCtrl->addError("External browser does not exist. Check settings in this source file.");
		}
	}
	
	private function isStringSurroundedWithQuotes($str) {
		$firstChar = substr($str, 0, 1);
		$lastChar = substr($str, strlen($str) - 1, 1);
		$isStringSurroundedWithQuotes =
			$this->charIsQuote($firstChar) &
			$this->charIsQuote($lastChar);

		return $isStringSurroundedWithQuotes;
	}
	
	private function charIsQuote($char) {
		return strcmp($char, '"') == 0;
	}
	
	private function removeStartAndEndCharFrom($str) {
		return substr($str, 1, strlen($str) - 2);
	}

	public function showImageBrowser() {
		if ($this->errorCtrl->isErrors()) {
			$this->errorOccured();
		} else {
			$this->htmlPageBuilder->setCss($this->getCss());
			echo $this->htmlPageBuilder->getHtmlPage();
		}
	}
	
	private function getCss() {
		$css = "";
		$css .= 'a';
		$css .= '{';
		$css .= '	font-size:24px; ';
		$css .= '}';
		
		return $css;
	}
	
	private function addBreak() {
		$this->htmlPageBuilder->addHtml("<hr style='color:#000000; background-color:#000000; height:2px' />");
	}
	
	public function isClient() {
		return $this->isClient === "true" ? true : false ;
	}
	
	public function showImageBrowserOnHostScreen() {
		if ($this->errorCtrl->isErrors()) {
			$this->errorOccured();
		} else {
			// Open browser that displays the html output
			$serverName = escapeshellarg($_SERVER['SERVER_NAME']);
			$phpSelf = escapeshellarg($_SERVER['PHP_SELF']);
			$imgDir = escapeshellarg($this->imgDir);
			$currentImageNum = escapeshellarg($this->currentImageNum);
			$htmlUrlHttp = "http://" . $serverName . $phpSelf . "?imgDir=" . $imgDir . "&imgNum=" . $currentImageNum . "&isClient=false";
			
			$cmd = "call " . Config::FILEPATH_TO_BROWSER() . " \"" . $htmlUrlHttp . "\"";
			$this->execInBackground($cmd);
		}
	}
	
	// Non-blocking execution of a command.
	private function execInBackground($cmd) { 
		if ($this->runningOsIsWindows()) {
			pclose(popen("start /B ". $cmd, "r"));  
		} else { 
			exec($cmd . " > /dev/null &");   
		} 
	}
	
	private function runningOsIsWindows() {
		// Alternative method: if (substr(php_uname(), 0, 7) == "Windows"){ 
		if (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN') {
			return true;
		} else {
			return false;
		}
	}
	
	private function errorOccured() {
		echo $this->errorCtrl->getErrorsAsHtml();
	}	
	
}

function getParam($requestArr, $param) {
	if (empty($requestArr[$param])) {
		return false;
	}
	
	return $requestArr[$param];
}

$errorCtrl = ErrorController::getInstance();

if (($imgNum = getParam($_GET, "imgNum")) === false) {
	$errorCtrl->addError("Missing parameter for image number");
}

if (($imgDir = getParam($_GET, "imgDir")) === false) {
	$errorCtrl->addError("Missing parameter for image directory");
}

if (($isClient = getParam($_GET, "isClient")) === false) {
	$errorCtrl->addError("Missing parameter for client/server setting");
}

if ($errorCtrl->isErrors()) {
	echo $errorCtrl->getErrorsAsHtml();
	return;
}

$imageBrowser = new ImageBrowser($imgDir, $imgNum, $isClient);

if ($errorCtrl->isErrors()) {
	echo $errorCtrl->getErrorsAsHtml();
	return;
}

$imageBrowser->showImageBrowser();

if ($errorCtrl->isErrors()) {
	echo $errorCtrl->getErrorsAsHtml();
	return;
}

// If the runner of this file is a client, show this page in a browser on the server as well.
if ($imageBrowser->isClient()) {
	$imageBrowser->showImageBrowserOnHostScreen();
}