<?php
/*
 * ------------------------------------------------------------------------------------------------
 * DESCRIPTION
 * Converts PPT slides to one HTML+JPG pair per slide.
 * ------------------------------------------------------------------------------------------------
 * HOW TO RUN:
 * $pptConverter = new PptConverter();
 * $outputImageDirectory = $pptConverter->convertPptToImages("http://yngvark.com/files/MSIS.ppt");
 * ------------------------------------------------------------------------------------------------
 * REQUIREMENTS/INSTALLATION:
 * - INSTALL OPENOFFICE: OpenOffice must be installed.
 *
 * - MACRO DEPENDENCY: Macro for convertion must be provided
 * (should come along this file), currently called ppt2html.odt.
 *
 * - CONFIG OPENOFFICE: OpenOffice must be configured to "Medium security"
 * for macros. Do this by opening OpenOffice, go Tools -> Options ->
 * OpenOffice.org (in left menu) -> Security -> Macro Security
 * (button on right panel) -> Medium (radio button).
 *
 * - CONFIG OPENOFFICE: OpenOffice must "trust" your macro. Do this by
 * following the same description as above, but switch to tab "Trusted
 * Sources" in stead of tab "Security level". Press button "Add..." at the
 * bottom, and add the directory where the ppt2html.odt macro is located.
 * ------------------------------------------------------------------------------------------------
 *
 * Yngvar Kristiansen, 11.05.2010
 *
*/
require_once('Config.php');
require_once('ErrorController.php');

class PptConverter {
	private $errorCtrl;
	
	public function PptConverter () {
		$this->errorCtrl = ErrorController::getInstance();
	}
	
	public function convertPptToImages($urlToPpt) {
		if (! $this->isValidUrl($urlToPpt)) {
			$this->errorCtrl->addError("Invalid URL.");
		}
		
		if (! $this->runningOsIsWindows()) {
			$this->errorCtrl->addError("Only Windows is currently supported by this application.");
		}
		
		if (! $this->errorCtrl->isErrors()) {
			return $this->convertPptToImagesInternal($urlToPpt);
		}
	}
	
	private function runningOsIsWindows() {
		// Alternative method: if (substr(php_uname(), 0, 7) == "Windows"){ 
		if (strtoupper(substr(PHP_OS, 0, 3)) === 'WIN') {
			return true;
		} else {
			return false;
		}
	}
	
	// TODO: Bettering naming of this function? I'm sceptical to the "Internal".
	private function convertPptToImagesInternal($urlToPpt) {
		$outputDirRelative = $this->createRndDir();
		// Converting to images actually implies converting each slide to a HTML/JPG pair,
		// because that is how the OpenOffice converter works. (We only need the JPG images,
		// however.)
		$outputHtmlAbsolute = getcwd() . "\\" . $outputDirRelative . "\\out.html";
		$ooMacroWithOutFileType = 
		$exePre =
			'call ' . Config::FILEPATH_OPENOFFICE() . ' ' .
			Config::FILEPATH_OPENOFFICE_CONVERT_MACRO() . ' ' .
			'macro://' . Config::OPENOFFICE_CONVERT_MACRO() .
			'/Standard.Convert.ConvertPPTtoHTMLAndExit(';
		$args = escapeshellarg($urlToPpt);
		$exePost = ',' . $outputHtmlAbsolute . ')'; // It is very important not to have space after
													// comma between source/PPT URL and
													// destination/HTML URL, or else the OpenOffice
													// macro won't run.

		$cmd = $exePre . $args . $exePost;
		exec($cmd);
		
		return $outputDirRelative;
	}
	
	// @return String The name of the created directory.
	private function createRndDir() {
		do {
			$rndString = "out_" . $this->getRndString(10);
		} while (is_dir($rndString));
		
		if (mkdir($rndString)) {
			return $rndString;
		} else {
			return false;
		}
	}

	private function getRndString($length = 32, $chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890')
	{
		$chars_length = (strlen($chars) - 1);
		$string = $chars{rand(0, $chars_length)};
		for ($i = 1; $i < $length; $i = strlen($string))
		{
			$r = $chars{rand(0, $chars_length)};
			if ($r != $string{$i - 1}) $string .=  $r;
		}
		
		return $string;
	}
	
	private function isValidUrl($url) { 
		if (! is_string($url)) {
			return false;
		}
		if (strlen($url) <= 0) {
			return false;
		}	
		
		$url = @parse_url($url); 

		if (!$url) { 
			return false; 
		} 

		$url = array_map('trim', $url); 
		$url['port'] = (!isset($url['port'])) ? 80 : (int)$url['port']; 

		$path = (isset($url['path'])) ? $url['path'] : '/'; 
		$path .= (isset($url['query'])) ? "?$url[query]" : ''; 

		if (isset($url['host']) AND $url['host'] != gethostbyname($url['host'])) { 
			if (PHP_VERSION >= 5) { 
				$headers = implode('', get_headers("$url[scheme]://$url[host]:$url[port]$path")); 
			} else { 
				$fp = fsockopen($url['host'], $url['port'], $errno, $errstr, 30); 
				
				if (!$fp) {
					return false; 
				}
				fputs($fp, "HEAD $path HTTP/1.1\r\nHost: $url[host]\r\n\r\n"); 
				$headers = fread($fp, 4096); 
				fclose($fp); 
			} 
			return (bool) preg_match('#^HTTP/.*\s+[(200|301|302)]+\s#i', $headers); 
		} 
		return false;
	}
}